# mypy: allow-untyped-defs
import functools
from collections import deque

import torch
from torch.utils._ordered_set import OrderedSet
from torch.utils._pytree import tree_map

from ..._dynamo.utils import counters
from ..ir import (
    ComputedBuffer,
    FixedLayout,
    FlexibleLayout,
    InputBuffer,
    StorageBox,
    Subgraph,
    TensorBox,
)
from ..lowering import lowerings
from ..pattern_matcher import (
    Arg,
    CallFunction,
    Match,
    PatternMatcherPass,
    register_graph_pattern,
)
from ..select_algorithm import (
    autotune_select_algorithm,
    ExternKernelChoice,
    SymbolicGridFn,
    TritonTemplate,
    TritonTemplateCaller,
)
from ..utils import ceildiv


B2B_GEMM_PASS = PatternMatcherPass(
    pass_name="b2b_gemm_pass",
)


@SymbolicGridFn
def b2b_gemm_grid(M, P, meta, *, cdiv):
    return (cdiv(M, meta["BLOCK_SIZE_M"]) * cdiv(P, meta["BLOCK_SIZE_P"]), 1, 1)


b2b_gemm_left_template = TritonTemplate(
    name="b2b_gemm_left",
    grid=b2b_gemm_grid,
    debug=False,
    source=r"""
{{def_kernel("A", "B", "C")}}


    # B2B_GEMM_LEFT_TRITON_ENTRANCE

    # dynamic shapes
    M = {{size("A", 0)}}
    N = {{size("A", 1)}}
    O = {{size("C", 0)}}
    P = {{size("C", 1)}}

    # dynamic strides
    stride_am = {{stride("A", 0)}}
    stride_an = {{stride("A", 1)}}
    stride_bn = {{stride("B", 0)}}
    stride_bo = {{stride("B", 1)}}
    stride_co = {{stride("C", 0)}}
    stride_cp = {{stride("C", 1)}}

    # output block counts
    num_m_block = tl.cdiv(M, BLOCK_SIZE_M)
    num_p_block = tl.cdiv(P, BLOCK_SIZE_P)

    # internal block counts
    num_n_block = tl.cdiv(N, BLOCK_SIZE_N)
    num_o_block = tl.cdiv(O, BLOCK_SIZE_O)

    # output block ids
    pid = tl.program_id(axis=0)
    m_block_id = pid // num_p_block
    p_block_id = pid % num_p_block

    # accumulator
    acc = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_P), dtype=tl.float32)

    # main loop
    offs_m = (m_block_id * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M))
    offs_p = (p_block_id * BLOCK_SIZE_P + tl.arange(0, BLOCK_SIZE_P))
    # (subgraph(A @ B) @ C)
    offs_o = tl.arange(0, BLOCK_SIZE_O)
    for _ in range(num_o_block):
        c_mask = (offs_o[:, None] < O) & (offs_p[None, :] < P)
        c_ptrs = C + (offs_o[:, None] * stride_co + offs_p[None, :] * stride_cp)
        c = tl.load(c_ptrs, mask=c_mask, other=0.0).to(tl.float32)  # BLOCK_SIZE_O * BLOCK_SIZE_P
        acc_ab = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_O), dtype=tl.float32)
        offs_n = tl.arange(0, BLOCK_SIZE_N)
        for __ in range(num_n_block):
            a_mask = (offs_m[:, None] < M) & (offs_n[None, :] < N)
            a_ptrs = A + (offs_m[:, None] * stride_am + offs_n[None, :] * stride_an)
            a = tl.load(a_ptrs, mask=a_mask, other=0.0).to(tl.float32)  # BLOCK_SIZE_M * BLOCK_SIZE_N
            b_mask = (offs_n[:, None] < N) & (offs_o[None, :] < O)
            b_ptrs = B + (offs_n[:, None] * stride_bn + offs_o[None, :] * stride_bo)
            b = tl.load(b_ptrs, mask=b_mask, other=0.0).to(tl.float32)  # BLOCK_SIZE_N * BLOCK_SIZE_O
            acc_ab += tl.dot(a, b, out_dtype=tl.float32)
            offs_n += BLOCK_SIZE_N
        # apply the subgraph
        {{ modification(
            subgraph_number=0,
            output_name="post_subgraph_acc_ab",
            inner_mm="acc_ab"
        ) | indent_except_first(2) }}
        acc += tl.dot(post_subgraph_acc_ab, c, out_dtype=tl.float32)
        offs_o += BLOCK_SIZE_O

    # type conversion
    acc = acc.to(tl.float16)

    # store preparation
    idx_m = offs_m[:, None]
    idx_p = offs_p[None, :]
    out_mask = (idx_m < M) & (idx_p < P)

    {{store_output(("idx_m", "idx_p"), "acc", "out_mask", val_shape=("BLOCK_SIZE_M", "BLOCK_SIZE_P"))}}
""",
)


b2b_gemm_right_template = TritonTemplate(
    name="b2b_gemm_right",
    grid=b2b_gemm_grid,
    debug=False,
    source=r"""
{{def_kernel("A", "B", "C")}}


    # B2B_GEMM_RIGHT_TRITON_ENTRANCE

    # dynamic shapes
    M = {{size("A", 0)}}
    N = {{size("A", 1)}}
    O = {{size("C", 0)}}
    P = {{size("C", 1)}}

    # dynamic strides
    stride_am = {{stride("A", 0)}}
    stride_an = {{stride("A", 1)}}
    stride_bn = {{stride("B", 0)}}
    stride_bo = {{stride("B", 1)}}
    stride_co = {{stride("C", 0)}}
    stride_cp = {{stride("C", 1)}}

    # output block counts
    num_m_block = tl.cdiv(M, BLOCK_SIZE_M)
    num_p_block = tl.cdiv(P, BLOCK_SIZE_P)

    # internal block counts
    num_n_block = tl.cdiv(N, BLOCK_SIZE_N)
    num_o_block = tl.cdiv(O, BLOCK_SIZE_O)

    # output block ids
    pid = tl.program_id(axis=0)
    m_block_id = pid // num_p_block
    p_block_id = pid % num_p_block

    # accumulator
    acc = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_P), dtype=tl.float32)

    # main loop (two cases)
    offs_m = (m_block_id * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M))
    offs_p = (p_block_id * BLOCK_SIZE_P + tl.arange(0, BLOCK_SIZE_P))
    # (A @ subgraph(B @ C))
    offs_n = tl.arange(0, BLOCK_SIZE_N)
    for _ in range(num_n_block):
        a_mask = (offs_m[:, None] < M) & (offs_n[None, :] < N)
        a_ptrs = A + (offs_m[:, None] * stride_am + offs_n[None, :] * stride_an)
        a = tl.load(a_ptrs, mask=a_mask, other=0.0).to(tl.float32)  # BLOCK_SIZE_M * BLOCK_SIZE_N
        acc_bc = tl.zeros((BLOCK_SIZE_N, BLOCK_SIZE_P), dtype=tl.float32)
        offs_o = tl.arange(0, BLOCK_SIZE_O)
        for __ in range(num_o_block):
            b_mask = (offs_n[:, None] < N) & (offs_o[None, :] < O)
            b_ptrs = B + (offs_n[:, None] * stride_bn + offs_o[None, :] * stride_bo)
            b = tl.load(b_ptrs, mask=b_mask, other=0.0).to(tl.float32)  # BLOCK_SIZE_N * BLOCK_SIZE_O
            c_mask = (offs_o[:, None] < O) & (offs_p[None, :] < P)
            c_ptrs = C + (offs_o[:, None] * stride_co + offs_p[None, :] * stride_cp)
            c = tl.load(c_ptrs, mask=c_mask, other=0.0).to(tl.float32)  # BLOCK_SIZE_O * BLOCK_SIZE_P
            acc_bc += tl.dot(b, c, out_dtype=tl.float32)
            offs_o += BLOCK_SIZE_O
        # apply the subgraph
        {{ modification(
            subgraph_number=0,
            output_name="post_subgraph_acc_bc",
            inner_mm="acc_bc"
        ) | indent_except_first(2) }}
        acc += tl.dot(a, post_subgraph_acc_bc, out_dtype=tl.float32)
        offs_n += BLOCK_SIZE_N

    # type conversion
    acc = acc.to(tl.float16)

    # store preparation
    idx_m = offs_m[:, None]
    idx_p = offs_p[None, :]
    out_mask = (idx_m < M) & (idx_p < P)

    {{store_output(("idx_m", "idx_p"), "acc", "out_mask", val_shape=("BLOCK_SIZE_M", "BLOCK_SIZE_P"))}}
""",
)


# Note: load_ratio_left and load_ratio_right are only calculating numbers
# in the trivial subgraph case; i.e. (A @ (B @ C)) or ((A @ B) @ C)


def load_ratio_left(
    M: int, N: int, O: int, P: int, m: int, n: int, o: int, p: int
) -> float:
    """
    compute the ratio of estimated numbers of loads in baseline and b2bgemm
    M, N, O, P are matrix sizes
    m, n, o, p are block sizes
    |       | baseline (lower bound)        | b2bgemm
    | load  | M * N + N * O + M * O + O * P | M / m * P / p * O / o * (o * p + N / n * (m * n + n * o))
    | store | M * O + M * P                 | M * P
    b2bgemm is always better on stores, but for loads we need to find out beneficial cases using this function
    """
    base = M * N + N * O + M * O + O * P
    gemm = (
        ceildiv(M, m)
        * ceildiv(P, p)
        * ceildiv(O, o)
        * (o * p + ceildiv(N, n) * (m * n + n * o))
    )
    return base / gemm


def load_ratio_right(
    M: int, N: int, O: int, P: int, m: int, n: int, o: int, p: int
) -> float:
    """
    compute the ratio of estimated numbers of loads in baseline and b2bgemm
    M, N, O, P are matrix sizes
    m, n, o, p are block sizes
    |       | baseline (lower bound)        | b2bgemm
    | load  | N * O + O * P + M * N + N * P | M / m * P / p * N / n * (m * n + O / o * (n * o + o * p))
    | store | N * P + M * P                 | M * P
    b2bgemm is always better on stores, but for loads we need to find out beneficial cases using this function
    """
    base = N * O + O * P + M * N + N * P
    gemm = (
        ceildiv(M, m)
        * ceildiv(P, p)
        * ceildiv(N, n)
        * (m * n + ceildiv(O, o) * (n * o + o * p))
    )
    return base / gemm


# the block sizes are limited by hardware (the shared memory)
# intuitively, the optimization works when the intermediate matrix is large
# and we assign large block sizes to large dimensions
b2b_gemm_configs = [
    {
        "BLOCK_SIZE_M": 128,
        "BLOCK_SIZE_N": 16,
        "BLOCK_SIZE_O": 16,
        "BLOCK_SIZE_P": 16,
        "num_stages": 4,
        "num_warps": 8,
    },
    {
        "BLOCK_SIZE_M": 128,
        "BLOCK_SIZE_N": 32,
        "BLOCK_SIZE_O": 32,
        "BLOCK_SIZE_P": 32,
        "num_stages": 2,
        "num_warps": 4,
    },
    {
        "BLOCK_SIZE_M": 128,
        "BLOCK_SIZE_N": 64,
        "BLOCK_SIZE_O": 64,
        "BLOCK_SIZE_P": 64,
        "num_stages": 2,
        "num_warps": 4,
    },
    {
        "BLOCK_SIZE_M": 128,
        "BLOCK_SIZE_N": 16,
        "BLOCK_SIZE_O": 128,
        "BLOCK_SIZE_P": 16,
        "num_stages": 4,
        "num_warps": 8,
    },
    {
        "BLOCK_SIZE_M": 128,
        "BLOCK_SIZE_N": 32,
        "BLOCK_SIZE_O": 128,
        "BLOCK_SIZE_P": 32,
        "num_stages": 2,
        "num_warps": 4,
    },
    {
        "BLOCK_SIZE_M": 128,
        "BLOCK_SIZE_N": 64,
        "BLOCK_SIZE_O": 128,
        "BLOCK_SIZE_P": 64,
        "num_stages": 2,
        "num_warps": 4,
    },
    {
        "BLOCK_SIZE_M": 16,
        "BLOCK_SIZE_N": 16,
        "BLOCK_SIZE_O": 16,
        "BLOCK_SIZE_P": 128,
        "num_stages": 4,
        "num_warps": 8,
    },
    {
        "BLOCK_SIZE_M": 32,
        "BLOCK_SIZE_N": 32,
        "BLOCK_SIZE_O": 32,
        "BLOCK_SIZE_P": 128,
        "num_stages": 2,
        "num_warps": 4,
    },
    {
        "BLOCK_SIZE_M": 64,
        "BLOCK_SIZE_N": 64,
        "BLOCK_SIZE_O": 64,
        "BLOCK_SIZE_P": 128,
        "num_stages": 2,
        "num_warps": 4,
    },
    {
        "BLOCK_SIZE_M": 16,
        "BLOCK_SIZE_N": 128,
        "BLOCK_SIZE_O": 16,
        "BLOCK_SIZE_P": 128,
        "num_stages": 4,
        "num_warps": 8,
    },
    {
        "BLOCK_SIZE_M": 32,
        "BLOCK_SIZE_N": 128,
        "BLOCK_SIZE_O": 32,
        "BLOCK_SIZE_P": 128,
        "num_stages": 2,
        "num_warps": 4,
    },
    {
        "BLOCK_SIZE_M": 64,
        "BLOCK_SIZE_N": 128,
        "BLOCK_SIZE_O": 64,
        "BLOCK_SIZE_P": 128,
        "num_stages": 2,
        "num_warps": 4,
    },
]


def is_b2b_gemm_good_on(
    is_left_assoc: bool,
    A_node: torch.fx.Node,
    B_node: torch.fx.Node,
    C_node: torch.fx.Node,
) -> bool:
    """
    checks whether the sizes are good for b2b_gemm
    """
    # basic checks
    if not all(["val" in A_node.meta, "val" in B_node.meta, "val" in C_node.meta]):
        return False
    fake_tensors = (
        A_node.meta["val"],
        B_node.meta["val"],
        C_node.meta["val"],
    )  # torch._subclasses.fake_tensor.FakeTensor

    A, B, C = fake_tensors

    def check_all_attr_true(objects, attr):
        return all(hasattr(obj, attr) and getattr(obj, attr) for obj in objects)

    if not check_all_attr_true(fake_tensors, "is_cuda") and not check_all_attr_true(
        fake_tensors, "is_xpu"
    ):
        return False
    if not all([len(A.shape) == 2, len(B.shape) == 2, len(C.shape) == 2]):
        return False
    if not ((A.shape[1] == B.shape[0]) and (B.shape[1] == C.shape[0])):
        return False
    # size checks: we only dispatch to B2B-GEMM when the average load ratio is > 1
    M, N = A.shape
    O, P = C.shape
    ratios = []
    if is_left_assoc:
        for config in b2b_gemm_configs:
            ratio = load_ratio_left(
                M,
                N,
                O,
                P,
                config["BLOCK_SIZE_M"],
                config["BLOCK_SIZE_N"],
                config["BLOCK_SIZE_O"],
                config["BLOCK_SIZE_P"],
            )
            ratios.append(ratio)
    else:
        for config in b2b_gemm_configs:
            ratio = load_ratio_right(
                M,
                N,
                O,
                P,
                config["BLOCK_SIZE_M"],
                config["BLOCK_SIZE_N"],
                config["BLOCK_SIZE_O"],
                config["BLOCK_SIZE_P"],
            )
            ratios.append(ratio)
    ratios.sort(reverse=True)
    average_ratio = 1.0
    for r in ratios[:3]:  # top 3 choices
        average_ratio *= r
    average_ratio = average_ratio ** (1 / 3)
    return (
        average_ratio > 1
    )  # even if average_ratio is close to 1, the number of stores is always better


def unoptimized_b2b_gemm(
    is_left_assoc: bool,
    subgraph: Subgraph,
    A: torch.Tensor,
    B: torch.Tensor,
    C: torch.Tensor,
    *,
    out: torch.Tensor,
) -> torch.Tensor:
    """
    The unoptimized version is used as a fallback when the b2b_gemm kernel is not beneficial.
    """
    if is_left_assoc:
        torch.mm(subgraph.graph_module(torch.mm(A, B)), C, out=out)
    else:
        torch.mm(A, subgraph.graph_module(torch.mm(B, C)), out=out)
    return out


unoptimized_choice = ExternKernelChoice(unoptimized_b2b_gemm)


def build_subgraph_buffer(
    args: list[TensorBox],
    subgraph: Subgraph,
):
    """
    This function is adapted from ../kernel/flex_attention.py.
    The goal is to take in the required args and produce the subgraph buffer
    The subgraph buffer is a ComputedBuffer that will be inlined into the triton template

    Args:
        args: The args that are passed into the subgraph
        subgraph: The Subgraph ir for which to produce the output node
    """
    cnt = 0
    env = {}
    for node in subgraph.graph_module.graph.nodes:
        if node.op == "placeholder":
            env[node] = args[cnt]
            cnt += 1
        elif node.op == "call_function":
            # For call_function we use the default lowerings and pass in the
            # already created TensorBoxes as args
            args, kwargs = tree_map(lambda x: env.get(x, x), (node.args, node.kwargs))
            env[node] = lowerings[node.target](*args, **kwargs)
        elif node.op == "output":

            def convert_output_node_to_buffer(output):
                if output is None:
                    return None
                output_node = output
                output_buffer = env[output_node]
                assert isinstance(output_buffer, TensorBox), (
                    "The output node for B2B-GEMM's subgraph must be a TensorBox, but got: ",
                    type(output_buffer),
                )
                assert isinstance(output_buffer.data, StorageBox), (
                    "The output node for B2B-GEMM's subgraph must be a StorageBox, but got: ",
                    type(output_buffer),
                )
                device = output_buffer.data.get_device()
                assert device is not None
                subgraph_buffer = ComputedBuffer(
                    name=None,
                    layout=FlexibleLayout(
                        device=device,
                        dtype=output_buffer.data.get_dtype(),
                        size=output_buffer.data.get_size(),
                    ),
                    data=output_buffer.data.data,  # type: ignore[arg-type]
                )
                return subgraph_buffer

            # node.args[0] should be a single element representing the output of the subgraph
            return tree_map(convert_output_node_to_buffer, node.args[0])

    raise ValueError("B2B-GEMM was passed a subgraph with no output node!")


def create_placeholder(
    name: str, dtype: torch.dtype, device: torch.device
) -> TensorBox:
    """
    Creates a placeholder input buffers for producing subgraph_output
    """
    input_buffer = InputBuffer(name=name, layout=FixedLayout(device, dtype, [], []))
    return TensorBox.create(input_buffer)


def tuned_b2b_gemm(
    is_left_assoc: bool,
    subgraph: Subgraph,
    A: torch._inductor.ir.TensorBox,
    B: torch._inductor.ir.TensorBox,
    C: torch._inductor.ir.TensorBox,
    *,
    layout=None,
) -> torch._inductor.ir.TensorBox:
    # call .realize() to get rid of Pointwise
    A.realize()
    B.realize()
    C.realize()
    layout = FixedLayout(
        A.get_device_or_error(),
        A.get_dtype(),
        [A.shape[0], C.shape[1]],  # type: ignore[index]
    )
    placeholders = [
        create_placeholder("inner_mm", A.get_dtype(), A.get_device_or_error())
    ]
    subgraph_buffer = build_subgraph_buffer(
        placeholders,  # type: ignore[arg-type, list-item]
        subgraph,
    )
    choices: list[TritonTemplateCaller] = []
    for config in b2b_gemm_configs:
        if is_left_assoc:
            b2b_gemm_left_template.maybe_append_choice(
                choices,
                input_nodes=(A, B, C),
                layout=layout,
                subgraphs=[subgraph_buffer],
                **config,
            )
        else:
            b2b_gemm_right_template.maybe_append_choice(
                choices,
                input_nodes=(A, B, C),
                layout=layout,
                subgraphs=[subgraph_buffer],
                **config,
            )
    # add the unoptimized choice to mitigate performance degradation
    choices.append(
        unoptimized_choice.bind(
            (A, B, C), layout, is_left_assoc=is_left_assoc, subgraph=subgraph
        )
    )
    # autotune
    node, _ = autotune_select_algorithm("b2b_gemm", choices, [A, B, C], layout)
    return node


# match the inner mm of a potential b2b_gemm
@register_graph_pattern(
    CallFunction(torch.ops.aten.mm, Arg(), Arg()),
    # pyrefly: ignore [bad-argument-type]
    pass_dict=B2B_GEMM_PASS,
)
def b2b_gemm_handler(match: Match, mat1: torch.fx.Node, mat2: torch.fx.Node) -> None:
    # match.args: list[torch.fx.Node]

    def is_pointwise_node(node: torch.fx.Node) -> bool:
        return (
            node.op == "call_function"
            and isinstance(node.target, torch._ops.OpOverload)
            and (torch.Tag.pointwise in node.target.tags)
        )

    def is_mm(node: torch.fx.Node) -> bool:
        return node.target is torch.ops.aten.mm.default

    # the inner MM
    inner_mm = match.nodes[-1]

    # find the (candidate) outer MM, which will be re-checked below to ensure every path reaches it
    # In a real (A @ f(B @ C)), every path starting from (B @ C) must reach (A @ _).
    outer_mm = None
    node = inner_mm
    while len(node.users) > 0:
        node = next(iter(node.users))
        if is_mm(node):
            outer_mm = node
            break
        elif is_pointwise_node(node):
            continue
        else:
            break
    if not outer_mm:
        return

    # find the unique input node for outer_mm representing f(B @ C) in (A @ f(B @ C))
    # we call it the "f_node"
    # when the pattern is simply (A @ (B @ C)), f_node is just inner_mm
    f_node = inner_mm
    while next(iter(f_node.users)) is not outer_mm:
        f_node = next(iter(f_node.users))

    def all_reach_via_pointwise_with_no_other_inputs(
        src: torch.fx.Node,
        dst: torch.fx.Node,
    ) -> tuple[bool, OrderedSet[torch.fx.Node]]:
        """
        check whether every user path from src reaches dst via pointwise nodes,
        with no other input nodes for the intermediates and dst;
        return
        (1) the Boolean value
        (2) the subgraph node set including src and dst (which only makes sense when the Boolean value is True)
        """
        visited = OrderedSet[torch.fx.Node]()
        input_counter: dict[torch.fx.Node, int] = {}

        all_reachable = True
        queue = deque([src])
        while queue:
            node = queue.popleft()
            if node not in visited:
                if node is dst:
                    visited.add(node)
                elif (node is src) or is_pointwise_node(node):
                    for user in node.users:
                        # for nodes other than dst, bookkeep their users' input counts
                        if user not in input_counter:
                            input_counter[user] = len(user.all_input_nodes)
                        input_counter[user] -= 1
                        # continue BFS
                        queue.append(user)
                    visited.add(node)
                else:
                    all_reachable = False
                    break

        return (
            all_reachable and all(count == 0 for count in input_counter.values()),
            visited,
        )

    # check inner_mm reaches f_node on every user path via pointwise nodes with no outside input_nodes
    ok, subgraph_node_set = all_reach_via_pointwise_with_no_other_inputs(
        inner_mm, f_node
    )
    if not ok:
        return

    # check inner_mm's inputs and f_node's outputs
    if not (len(inner_mm.all_input_nodes) == 2 and len(f_node.users) == 1):
        return

    # at this point, the nodes between inner_mm and f_node (both included)
    # are all used internally inside (A @ subgraph(B @ C))
    # i.e. they neither have other users nor have other inputs

    # original graph and module
    graph, module = inner_mm.graph, inner_mm.graph.owning_module

    # construct the new (sub)graph
    subgraph_node_list: list[
        torch.fx.Node
    ] = []  # ordered list of nodes used for node removal later
    new_graph: torch.fx.Graph = torch.fx.Graph()
    node_remapping: dict[torch.fx.Node, torch.fx.Node] = {}
    new_input_anchor: torch.fx.Node  # inner_mm, to be changed to an input node
    new_output_anchor: torch.fx.Node  # f_node, to be used to construct an output node
    new_input_node: torch.fx.Node
    new_output_node: torch.fx.Node
    for node in graph.nodes:  # preserve the order of nodes
        if node in subgraph_node_set:
            subgraph_node_list.append(node)
            new_node = new_graph.node_copy(node, lambda x: node_remapping.get(x, x))
            node_remapping[node] = new_node
            if node is inner_mm:
                new_input_anchor = new_node
            if node is f_node:
                new_output_anchor = new_node
    # pyrefly: ignore [unbound-name]
    if new_input_anchor is not new_output_anchor:  # subgraph is non-trivial
        # update the input node
        # pyrefly: ignore [unbound-name]
        with new_graph.inserting_before(new_input_anchor):
            new_input_node = new_graph.placeholder(name="subgraph_input")
            # pyrefly: ignore [unbound-name]
            new_input_node.meta.update(new_input_anchor.meta)
            # pyrefly: ignore [unbound-name]
            new_input_anchor.replace_all_uses_with(new_input_node)
        # pyrefly: ignore [unbound-name]
        new_graph.erase_node(new_input_anchor)
        # add the output node
        # pyrefly: ignore [unbound-name]
        new_output_node = new_graph.output(new_output_anchor)
        # pyrefly: ignore [unbound-name]
        new_output_node.meta.update(new_output_anchor.meta)
    else:  # subgraph is trivial, e.g. (A @ (B @ C))
        # update the input node
        # pyrefly: ignore [unbound-name]
        with new_graph.inserting_before(new_input_anchor):
            new_input_node = new_graph.placeholder(name="subgraph_input")
            # pyrefly: ignore [unbound-name]
            new_input_node.meta.update(new_input_anchor.meta)
            # pyrefly: ignore [unbound-name]
            new_input_anchor.replace_all_uses_with(new_input_node)
        # pyrefly: ignore [unbound-name]
        new_graph.erase_node(new_input_anchor)
        # update the output node (don't use new_output_anchor since it has been erased)
        new_output_node = new_graph.output(new_input_node)
        new_output_node.meta.update(new_input_node.meta)
    new_graph.lint()

    # construct the subgraph
    subgraph = Subgraph(
        name="subgraph", graph_module=torch.fx.GraphModule(module, new_graph)
    )

    # two cases
    # (1) (subgraph(A @ B) @ C), called "left_assoc"
    # (2) (A @ subgraph(B @ C)), called "right_assoc"
    is_left_assoc = outer_mm.args[0] is f_node

    # find the nodes A, B, C and check the sizes
    A: torch.fx.Node
    B: torch.fx.Node
    C: torch.fx.Node
    if is_left_assoc:
        A = inner_mm.args[0]  # type: ignore[assignment]
        B = inner_mm.args[1]  # type: ignore[assignment]
        C = outer_mm.args[1]  # type: ignore[assignment]
    else:
        A = outer_mm.args[0]  # type: ignore[assignment]
        B = inner_mm.args[0]  # type: ignore[assignment]
        C = inner_mm.args[1]  # type: ignore[assignment]
    if not is_b2b_gemm_good_on(is_left_assoc, A, B, C):
        return

    # finally update the original graph
    counters["inductor"]["b2b_gemm"] += 1
    graph = match.graph
    with graph.inserting_before(outer_mm):
        function = functools.partial(tuned_b2b_gemm, is_left_assoc, subgraph)
        function.__name__ = tuned_b2b_gemm.__name__  # type: ignore[attr-defined]
        function._inductor_lowering_function = True  # type: ignore[attr-defined]
        replacement: torch.fx.Node = graph.call_function(
            function,
            (A, B, C),
            match.kwargs,
        )
        replacement.meta.update(outer_mm.meta)
        outer_mm.replace_all_uses_with(replacement)
    # erase unnecessary nodes
    graph.erase_node(outer_mm)
    for node in reversed(subgraph_node_list):
        graph.erase_node(node)
    graph.lint()



# Owner(s): ["module: inductor"]
from dataclasses import dataclass

from torch._inductor.remote_cache import (
    RemoteCache,
    RemoteCacheBackend,
    RemoteCachePassthroughSerde,
)
from torch.testing._internal.common_utils import TestCase


class FailingBackend(RemoteCacheBackend):
    def _get(self, key):
        raise AssertionError("testget")

    def _put(self, key, data):
        raise AssertionError("testput")


class NoopBackend(RemoteCacheBackend):
    def _get(self, key):
        return None

    def _put(self, key, data):
        return None


@dataclass
class TestSample:
    fail: str = None


class FakeCache(RemoteCache):
    def __init__(self):
        super().__init__(FailingBackend(), RemoteCachePassthroughSerde())

    def _create_sample(self):
        return TestSample()

    def _log_sample(self, sample):
        self.sample = sample


class TestRemoteCache(TestCase):
    def test_normal_logging(
        self,
    ) -> None:
        c = RemoteCache(NoopBackend(), RemoteCachePassthroughSerde())
        c.put("test", "value")
        c.get("test")

    def test_failure_no_sample(
        self,
    ) -> None:
        c = RemoteCache(FailingBackend(), RemoteCachePassthroughSerde())
        with self.assertRaises(AssertionError):
            c.put("test", "value")
        with self.assertRaises(AssertionError):
            c.get("test")

    def test_failure_logging(
        self,
    ) -> None:
        c = FakeCache()
        with self.assertRaises(AssertionError):
            c.put("test", "value")
        self.assertEqual(c.sample.fail_reason, "testput")
        with self.assertRaises(AssertionError):
            c.get("test")
        self.assertEqual(c.sample.fail_reason, "testget")


if __name__ == "__main__":
    from torch._inductor.test_case import run_tests

    run_tests()



# Owner(s): ["module: dynamo"]

import dataclasses
import types
import unittest

import torch
import torch._dynamo.testing as dynamo_testing
from torch._dynamo.test_case import run_tests, TestCase


class SlotsOnly:
    __slots__ = ("x", "y")

    def __init__(self, x, y):
        self.x = x
        self.y = y


class SlotsAndDict:
    __slots__ = ("x", "__dict__")

    def __init__(self, x):
        self.x = x


@dataclasses.dataclass(frozen=True, slots=True)
class FrozenSlots:
    x: int
    y: int


class SlotsAndSetattr:
    __slots__ = ("x",)

    def __init__(self, x):
        self.x = x

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value * 2)


class SlotsAndDictAndSetattr:
    __slots__ = ("x", "__dict__")

    def __init__(self, x):
        self.x = x

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value * 2)


class SlotsBase:
    __slots__ = ("x",)

    def __init__(self):
        self.x = 0


class SlotsDerived(SlotsBase):
    __slots__ = ("y",)

    def __init__(self):
        super().__init__()
        self.y = 0


class Plain:
    pass


class SlotsChildOfPlain(Plain):
    __slots__ = ("z",)

    def __init__(self):
        self.z = 0


class Slots:
    __slots__ = ("x",)


class SlotsShadowed(SlotsBase):
    x = 42  # class attribute shadows parent's slot descriptor


class SlotsAndProperty:
    __slots__ = ("_x",)

    def __init__(self, x):
        self._x = x

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        self._x = value * 2


class TestSlotsAttrAssignment(TestCase):
    """Tests for attribute assignment on objects with __slots__."""

    def test_valid_slot_assignment(self):
        # Case 1: assign to a declared slot — should succeed
        def fn(t):
            obj = SlotsOnly(1, 2)
            obj.x = 99
            return t + obj.x

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_invalid_slot_assignment_raises(self):
        # Case 2: assign to an undeclared attr on a slotted object (no __dict__)
        # should raise AttributeError in eager; compiled raises an exception too
        def fn(t):
            obj = SlotsOnly(1, 2)
            obj.z = 99
            return t + obj.x

        compiled_fn = torch.compile(fn, backend="eager", fullgraph=True)
        t = torch.ones(1)
        self.assertRaises(AttributeError, fn, t)
        self.assertRaises(Exception, compiled_fn, t)

    def test_slots_with_dict_allows_arbitrary_attrs(self):
        # Case 3: __slots__ includes __dict__ — arbitrary attr assignment should work
        def fn(t):
            obj = SlotsAndDict(1)
            obj.extra = 42
            return t + obj.x + obj.extra

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_frozen_dataclass_with_slots_construction(self):
        # Case 4: frozen dataclass with slots uses object.__setattr__ in __init__
        # to bypass the frozen __setattr__. Dynamo must allow this for slot descriptors.
        def fn(t):
            obj = FrozenSlots(3, 4)
            return t + obj.x + obj.y

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_custom_setattr_with_slots(self):
        # Case 5: __slots__ + custom __setattr__ — the custom __setattr__ is traced
        def fn(t):
            obj = SlotsAndSetattr(1)
            obj.x = 10
            return t + obj.x

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_with_dict_valid_slot_assignment(self):
        # Case 6: __slots__ + __dict__: assigning to a declared slot still works
        def fn(t):
            obj = SlotsAndDict(1)
            obj.x = 99
            return t + obj.x

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_with_dict_undeclared_attr_goes_to_dict(self):
        # Case 7: __slots__ + __dict__: assigning to an undeclared attr goes to
        # __dict__ instead of raising AttributeError
        def fn(t):
            obj = SlotsAndDict(1)
            obj.z = 42
            return t + obj.z

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_custom_setattr_with_slots_and_dict(self):
        # Case 8: __slots__ + __dict__ + custom __setattr__ — custom __setattr__
        # is traced for both slot and non-slot attrs
        def fn(t):
            obj = SlotsAndDictAndSetattr(1)
            obj.x = 10
            obj.extra = 3
            return t + obj.x + obj.extra

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_inheritance_parent_and_child_slots(self):
        # Subclass adds its own slot on top of parent's slot — both accessible
        def fn(t):
            obj = SlotsDerived()
            obj.x = 1
            obj.y = 2
            return t + obj.x + obj.y

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_child_inherits_dict_from_no_slots_parent(self):
        # Subclass with __slots__ inheriting from a parent without __slots__
        # gets __dict__ from the parent, so arbitrary attrs are allowed
        def fn(t):
            obj = SlotsChildOfPlain()
            obj.z = 1
            obj.extra = 42
            return t + obj.z + obj.extra

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_object_setattr_bypasses_custom_setattr(self):
        # object.__setattr__ skips the custom __setattr__ and writes directly to slot
        def fn(t):
            obj = SlotsAndSetattr(1)
            object.__setattr__(obj, "x", 5)
            return t + obj.x

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_getattr_default_on_unset_slot(self):
        # getattr with a default on an unset slot returns the default
        def fn(t):
            obj = Slots()
            val = getattr(obj, "x", 99)
            return t + val

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slot_read_after_delete_raises(self):
        # Reading a slot after deletion raises AttributeError in both eager and compiled
        def fn(t):
            obj = Slots()
            obj.x = 1
            del obj.x
            return t + obj.x

        compiled_fn = torch.compile(fn, backend="eager", fullgraph=True)
        t = torch.ones(1)
        self.assertRaises(AttributeError, fn, t)
        self.assertRaises(Exception, compiled_fn, t)

    def test_slot_shadowed_by_class_attribute(self):
        # Class attribute in subclass shadows parent slot descriptor:
        # reads return the class attribute, writes raise AttributeError
        def fn(t):
            obj = SlotsShadowed()
            return t + obj.x  # returns class attr 42

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slot_assignment_with_object_as_argument(self):
        # Slotted object passed as argument (not created inside fn)
        def fn(t, obj):
            obj.x = 10
            return t + obj.x

        compiled_fn = torch.compile(fn, backend="eager", fullgraph=True)
        t = torch.ones(1)
        obj = Slots()
        self.assertEqual(fn(t.clone(), obj), compiled_fn(t.clone(), obj))

    def test_slot_mutation_materialized_on_argument(self):
        # Slot mutation on an object passed as argument must be visible after
        # the compiled function returns (side effect materialization)
        def fn(t, obj):
            obj.x = 10
            return t.sin()

        compiled_fn = torch.compile(fn, backend="eager", fullgraph=True)
        obj = Slots()
        compiled_fn(torch.ones(1), obj)
        self.assertEqual(obj.x, 10)

    def test_slot_delete_materialized(self):
        # del on a slot inside a compiled fn must be visible after the call returns
        def fn(t, obj):
            del obj.x
            return t.sin()

        compiled_fn = torch.compile(fn, backend="eager", fullgraph=True)
        obj = Slots()
        obj.x = 1
        compiled_fn(torch.ones(1), obj)
        self.assertFalse(hasattr(obj, "x"))

    def test_hasattr_on_slotted_object(self):
        # hasattr inside compiled code reflects actual slot state
        def fn(t):
            obj = Slots()
            before = hasattr(obj, "x")  # False — slot not set
            obj.x = 5
            after = hasattr(obj, "x")  # True — slot is now set
            return t + before + after

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_with_property_setter(self):
        # property setter is called instead of writing directly to the slot
        def fn(t):
            obj = SlotsAndProperty(1)
            obj.x = 5  # calls setter: _x = 5 * 2 = 10
            return t + obj.x  # calls getter: returns _x = 10

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slot_assignment_no_recompile_same_type(self):
        # Calling compiled fn repeatedly with the same slotted object type
        # must not trigger recompilation
        cnts = dynamo_testing.CompileCounter()

        def fn(t, obj):
            obj.x = 10
            return t + obj.x

        compiled_fn = torch.compile(fn, backend=cnts)
        t = torch.ones(1)
        compiled_fn(t, Slots())
        compiled_fn(t, Slots())
        compiled_fn(t, Slots())
        self.assertEqual(cnts.frame_count, 1)

    def test_slot_assignment_recompiles_on_type_change(self):
        # Compiled fn sees slot assigned to int first, then float — guards recompile
        cnts = dynamo_testing.CompileCounter()

        def fn(t, a, obj):
            obj.x = a
            return t + obj.x

        compiled_fn = torch.compile(fn, backend=cnts)
        t = torch.ones(1)

        compiled_fn(t, 1, Slots())
        compiled_fn(t, 1, Slots())
        self.assertEqual(cnts.frame_count, 1)  # same type, no recompile

        x = t.clone()
        res = compiled_fn(x, 1.0, Slots())
        self.assertEqual(cnts.frame_count, 2)  # float instead of int — recompile
        self.assertEqual(res, fn(x, 1.0, Slots()))


class WithGetattribute:
    # __slots__ = ("x", "_side_effects")

    def __init__(self, x):
        object.__setattr__(self, "x", x)
        object.__setattr__(self, "_side_effects", set())

    def __getattribute__(self, name):
        effects = object.__getattribute__(self, "_side_effects")
        effects.add(name)
        return object.__getattribute__(self, name)


class TestSlotsFromCPython(TestCase):
    """Slot tests extracted from CPython's test_descr.py::test_slots."""

    def setUp(self):
        super().setUp()
        self._u_prev = torch._dynamo.config.enable_trace_unittest
        torch._dynamo.config.enable_trace_unittest = True

    def tearDown(self):
        super().tearDown()
        torch._dynamo.config.enable_trace_unittest = self._u_prev

    def test_slots_empty(self):
        class C:
            __slots__ = []

        def fn(t):
            x = C()
            self.assertFalse(hasattr(x, "__dict__"))
            self.assertFalse(hasattr(x, "foo"))
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_single(self):
        class C:
            __slots__ = ["a"]

        def fn(t):
            x = C()
            self.assertFalse(hasattr(x, "__dict__"))
            self.assertFalse(hasattr(x, "a"))
            x.a = 1
            self.assertEqual(x.a, 1)
            x.a = None
            self.assertEqual(x.a, None)
            del x.a
            self.assertFalse(hasattr(x, "a"))
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_multiple(self):
        class C:
            __slots__ = ["a", "b", "c"]

        def fn(t):
            x = C()
            self.assertFalse(hasattr(x, "__dict__"))
            self.assertFalse(hasattr(x, "a"))
            self.assertFalse(hasattr(x, "b"))
            self.assertFalse(hasattr(x, "c"))
            x.a = 1
            x.b = 2
            x.c = 3
            self.assertEqual(x.a, 1)
            self.assertEqual(x.b, 2)
            self.assertEqual(x.c, 3)
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_name_mangling(self):
        class C:
            __slots__ = ["__a"]

            def __init__(self, value):
                self.__a = value

            def get(self):
                return self.__a

        def fn(t):
            x = C(5)
            self.assertFalse(hasattr(x, "__dict__"))
            self.assertFalse(hasattr(x, "__a"))
            self.assertEqual(x.get(), 5)
            try:
                x.__a = 6
            except AttributeError:
                pass
            else:
                self.fail("Double underscored names not mangled")
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_string_not_expanded(self):
        # A single string is not expanded as a sequence
        class C:
            __slots__ = "abc"  # noqa: PLC0205

        def fn(t):
            c = C()
            c.abc = 5
            self.assertEqual(c.abc, 5)
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_tuple(self):
        slots = ("foo", "bar")

        class C:
            __slots__ = slots

        def fn(t):
            x = C()
            x.foo = 5
            self.assertEqual(x.foo, 5)
            self.assertIs(type(slots[0]), str)
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_get_unset_raises(self):
        class X:
            __slots__ = "a"  # noqa: PLC0205

        def fn(t):
            with self.assertRaises(AttributeError):
                X().a
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_str_subclass(self):
        # gh-98783: string subclass in __slots__
        class SubStr(str):  # noqa: SLOT000
            pass

        class X:
            __slots__ = (SubStr("x"),)

        def fn(t):
            X().x = 1
            with self.assertRaises(AttributeError):
                X().a
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_special_dict(self):
        # __dict__ in __slots__ enables arbitrary attr assignment
        class D:
            __slots__ = ["__dict__"]

        def fn(t):
            a = D()
            self.assertTrue(hasattr(a, "__dict__"))
            self.assertFalse(hasattr(a, "__weakref__"))
            a.foo = 42
            self.assertEqual(a.__dict__, {"foo": 42})
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_special_weakref(self):
        # __weakref__ in __slots__ — no __dict__, arbitrary attr raises
        class W:
            __slots__ = ["__weakref__"]

        def fn(t):
            a = W()
            self.assertTrue(hasattr(a, "__weakref__"))
            self.assertFalse(hasattr(a, "__dict__"))
            with self.assertRaises(AttributeError):
                a.foo = 42
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_special_inherit_dict_weakref(self):
        # Inheriting from both __dict__ and __weakref__ slot classes
        class D:
            __slots__ = ["__dict__"]

        class W:
            __slots__ = ["__weakref__"]

        class C1(W, D):
            __slots__ = []

        def fn(t):
            a = C1()
            self.assertTrue(hasattr(a, "__dict__"))
            self.assertTrue(hasattr(a, "__weakref__"))
            a.foo = 42
            self.assertEqual(a.__dict__, {"foo": 42})
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    @unittest.expectedFailure
    def test_slots_special2_classcell(self):
        # Testing __classcell__ in __slots__
        class Meta(type):
            def __new__(metacls, name, bases, namespace, attr):
                self.assertIn(attr, namespace)
                return super().__new__(metacls, name, bases, namespace)

        class C1:
            def __init__(self):
                self.b = 42

        class C2(C1, metaclass=Meta, attr="__classcell__"):
            __slots__ = ["__classcell__"]

            def __init__(self):
                super().__init__()

        def fn(t):
            self.assertIsInstance(
                C2.__dict__["__classcell__"], types.MemberDescriptorType
            )
            c = C2()
            self.assertEqual(c.b, 42)
            self.assertFalse(hasattr(c, "__classcell__"))
            c.__classcell__ = 42
            self.assertEqual(c.__classcell__, 42)
            with self.assertRaises(TypeError):

                class C3:
                    __classcell__ = 42
                    __slots__ = ["__classcell__"]

            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_slots_multiple_inheritance(self):
        # SF bug 575229: multiple inheritance w/ slots dumps core
        class A:
            __slots__ = ()

        class B:
            pass

        class C(A, B):
            __slots__ = ()

        def fn(t):
            self.assertTrue(hasattr(C, "__dict__"))
            self.assertTrue(hasattr(C, "__weakref__"))
            C().x = 2
            return t.sin()

        dynamo_testing.standard_test(self, fn, nargs=1)


class TestUserDefinedClassDict(TestCase):
    def test_class_dict_read(self):
        class MyClass:
            x = 3

        def fn(t):
            t = t + MyClass.__dict__["x"]
            t = t + MyClass.__dict__.get("x", 0)
            t = t + MyClass.__dict__.get("z", 99)
            t = t + (1 if "x" in MyClass.__dict__ else 0)
            t = t + (1 if "z" in MyClass.__dict__ else 0)
            return t

        dynamo_testing.standard_test(self, fn, nargs=1)

    def test_class_dict_via_arg(self):
        class MyClass:
            x = 7

        def fn(t, cls):
            return t + cls.__dict__.get("x", 0)

        cnt = dynamo_testing.CompileCounter()
        compiled = torch.compile(fn, backend=cnt)
        result = compiled(torch.tensor([0.0]), MyClass)
        self.assertEqual(result, torch.tensor([7.0]))

    def test_class_dict_mutation_recompiles(self):
        # Mutating a class attribute between calls should trigger recompilation,
        # and the compiled function should see the updated value.
        class MyClass:
            x = 1

        def fn(t):
            return t + MyClass.__dict__["x"]

        cnt = dynamo_testing.CompileCounter()
        compiled = torch.compile(fn, backend=cnt)

        result1 = compiled(torch.tensor([0.0]))
        self.assertEqual(result1, torch.tensor([1.0]))
        self.assertEqual(cnt.frame_count, 1)

        MyClass.x = 10
        result2 = compiled(torch.tensor([0.0]))
        self.assertEqual(result2, torch.tensor([10.0]))
        # Should have recompiled due to guard failure
        self.assertEqual(cnt.frame_count, 2)

    def test_class_dict_add_key_recompiles(self):
        # Adding a new attribute to the class should trigger recompilation
        # when the compiled code checks for key presence.
        class MyClass:
            x = 1

        def fn(t):
            return t + (1 if "y" in MyClass.__dict__ else 0)

        cnt = dynamo_testing.CompileCounter()
        compiled = torch.compile(fn, backend=cnt)

        result1 = compiled(torch.tensor([0.0]))
        self.assertEqual(result1, torch.tensor([0.0]))
        self.assertEqual(cnt.frame_count, 1)

        MyClass.y = 99
        result2 = compiled(torch.tensor([0.0]))
        self.assertEqual(result2, torch.tensor([1.0]))
        # Should have recompiled
        self.assertEqual(cnt.frame_count, 2)

    def test_class_dict_delete_key_recompiles(self):
        # Deleting a class attribute should trigger recompilation.
        class MyClass:
            x = 5
            y = 10

        def fn(t):
            return t + MyClass.__dict__.get("y", 0)

        cnt = dynamo_testing.CompileCounter()
        compiled = torch.compile(fn, backend=cnt)

        result1 = compiled(torch.tensor([0.0]))
        self.assertEqual(result1, torch.tensor([10.0]))
        self.assertEqual(cnt.frame_count, 1)

        del MyClass.y
        result2 = compiled(torch.tensor([0.0]))
        self.assertEqual(result2, torch.tensor([0.0]))
        # Should have recompiled
        self.assertEqual(cnt.frame_count, 2)


if __name__ == "__main__":
    run_tests()



# Owner(s): ["module: inductor"]

import functools
import logging

import torch
from torch._inductor.runtime.benchmarking import benchmarker
from torch._inductor.test_case import run_tests, TestCase
from torch._inductor.utils import do_bench_using_profiling


log = logging.getLogger(__name__)

device_type = (
    acc.type
    if (acc := torch.accelerator.current_accelerator(check_available=True))
    else "cpu"
)


class TestBench(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        x = torch.rand(1024, 10).to(device_type).half()
        w = torch.rand(512, 10).to(device_type).half()
        cls._bench_fn = functools.partial(torch.nn.functional.linear, x, w)

    def test_benchmarker(self):
        res = benchmarker.benchmark_gpu(self._bench_fn)
        log.warning("do_bench result: %s", res)
        self.assertGreater(res, 0)

    def test_do_bench_using_profiling(self):
        res = do_bench_using_profiling(self._bench_fn)
        log.warning("do_bench_using_profiling result: %s", res)
        self.assertGreater(res, 0)


if __name__ == "__main__":
    run_tests(device_type)



import torch.nn.functional as F
from torch import Tensor

from .module import Module


__all__ = ["PixelShuffle", "PixelUnshuffle"]


class PixelShuffle(Module):
    r"""Rearrange elements in a tensor according to an upscaling factor.

    Rearranges elements in a tensor of shape :math:`(*, C \times r^2, H, W)`
    to a tensor of shape :math:`(*, C, H \times r, W \times r)`, where r is an upscale factor.

    This is useful for implementing efficient sub-pixel convolution
    with a stride of :math:`1/r`.

    See the paper:
    `Real-Time Single Image and Video Super-Resolution Using an Efficient Sub-Pixel Convolutional Neural Network`_
    by Shi et al. (2016) for more details.

    Args:
        upscale_factor (int): factor to increase spatial resolution by

    Shape:
        - Input: :math:`(*, C_{in}, H_{in}, W_{in})`, where * is zero or more batch dimensions
        - Output: :math:`(*, C_{out}, H_{out}, W_{out})`, where

    .. math::
        C_{out} = C_{in} \div \text{upscale\_factor}^2

    .. math::
        H_{out} = H_{in} \times \text{upscale\_factor}

    .. math::
        W_{out} = W_{in} \times \text{upscale\_factor}

    Examples::

        >>> pixel_shuffle = nn.PixelShuffle(3)
        >>> input = torch.randn(1, 9, 4, 4)
        >>> output = pixel_shuffle(input)
        >>> print(output.size())
        torch.Size([1, 1, 12, 12])

    .. _Real-Time Single Image and Video Super-Resolution Using an Efficient Sub-Pixel Convolutional Neural Network:
        https://arxiv.org/abs/1609.05158
    """

    __constants__ = ["upscale_factor"]
    upscale_factor: int

    def __init__(self, upscale_factor: int) -> None:
        super().__init__()
        self.upscale_factor = upscale_factor

    def forward(self, input: Tensor) -> Tensor:
        """
        Runs the forward pass.
        """
        return F.pixel_shuffle(input, self.upscale_factor)

    def extra_repr(self) -> str:
        """
        Return the extra representation of the module.
        """
        return f"upscale_factor={self.upscale_factor}"


class PixelUnshuffle(Module):
    r"""Reverse the PixelShuffle operation.

    Reverses the :class:`~torch.nn.PixelShuffle` operation by rearranging elements
    in a tensor of shape :math:`(*, C, H \times r, W \times r)` to a tensor of shape
    :math:`(*, C \times r^2, H, W)`, where r is a downscale factor.

    See the paper:
    `Real-Time Single Image and Video Super-Resolution Using an Efficient Sub-Pixel Convolutional Neural Network`_
    by Shi et al. (2016) for more details.

    Args:
        downscale_factor (int): factor to decrease spatial resolution by

    Shape:
        - Input: :math:`(*, C_{in}, H_{in}, W_{in})`, where * is zero or more batch dimensions
        - Output: :math:`(*, C_{out}, H_{out}, W_{out})`, where

    .. math::
        C_{out} = C_{in} \times \text{downscale\_factor}^2

    .. math::
        H_{out} = H_{in} \div \text{downscale\_factor}

    .. math::
        W_{out} = W_{in} \div \text{downscale\_factor}

    Examples::

        >>> pixel_unshuffle = nn.PixelUnshuffle(3)
        >>> input = torch.randn(1, 1, 12, 12)
        >>> output = pixel_unshuffle(input)
        >>> print(output.size())
        torch.Size([1, 9, 4, 4])

    .. _Real-Time Single Image and Video Super-Resolution Using an Efficient Sub-Pixel Convolutional Neural Network:
        https://arxiv.org/abs/1609.05158
    """

    __constants__ = ["downscale_factor"]
    downscale_factor: int

    def __init__(self, downscale_factor: int) -> None:
        super().__init__()
        self.downscale_factor = downscale_factor

    def forward(self, input: Tensor) -> Tensor:
        """
        Runs the forward pass.
        """
        return F.pixel_unshuffle(input, self.downscale_factor)

    def extra_repr(self) -> str:
        """
        Return the extra representation of the module.
        """
        return f"downscale_factor={self.downscale_factor}"



# Owner(s): ["oncall: distributed"]

import copy

import torch.nn as nn
from torch.distributed.fsdp import FSDPModule, fully_shard
from torch.testing._internal.common_fsdp import FSDPTestMultiThread, MLP
from torch.testing._internal.common_utils import run_tests


class TestFullyShardState(FSDPTestMultiThread):
    @property
    def world_size(self) -> int:
        return 1

    def test_fully_shard_state(self):
        """
        Tests the ability to get the state object from a fully sharded module.
        """
        num_mlps = 3
        model = nn.Sequential(*[MLP(8) for _ in range(num_mlps)])
        for mlp in model:
            fully_shard(mlp)
        fully_shard(model)
        root_state = fully_shard.state(model)
        self.assertTrue(root_state is not None)
        all_states = [root_state] + [fully_shard.state(mlp) for mlp in model]
        # Check that each `fully_shard` call constructs a distinct state object
        self.assertEqual(len(set(all_states)), num_mlps + 1)

    def test_fully_shard_reapply(self):
        model = MLP(8)
        fully_shard(model)
        with self.assertRaisesRegex(
            AssertionError,
            "Each distinct composable distributed API can only be applied to a module once.",
        ):
            fully_shard(model)

    def test_fully_shard_cls(self):
        # Check that we only swap class for the module passed to `fully_shard`
        model = MLP(8)
        fully_shard(model)
        self.assertTrue(isinstance(model, MLP))
        self.assertTrue(isinstance(model, FSDPModule))
        self.assertEqual(model.__class__.__name__, "FSDPMLP")
        for module in model.modules():
            if module is model:
                continue
            self.assertFalse(isinstance(module, FSDPModule))

        # Check that slicing into a `Sequential` does not preserve FSDP
        model = nn.Sequential(*[MLP(8) for _ in range(3)])
        fully_shard(model)
        self.assertTrue(isinstance(model, nn.Sequential))
        self.assertTrue(isinstance(model, FSDPModule))
        self.assertEqual(model.__class__.__name__, "FSDPSequential")
        sliced_model = model[:2]
        self.assertTrue(isinstance(sliced_model, nn.Sequential))
        self.assertFalse(isinstance(sliced_model, FSDPModule))

    def test_fully_shard_unsupported_module_cls(self):
        regex = (
            r"fully\_shard does not support containers that do not implement forward"
        )
        model = nn.ModuleList([MLP(8) for _ in range(3)])
        with self.assertRaisesRegex(ValueError, regex):
            fully_shard(model)
        model = nn.ModuleDict({"1": MLP(8), "2": MLP(8)})
        with self.assertRaisesRegex(ValueError, regex):
            fully_shard(model)

    def test_fully_shard_deepcopy(self):
        model = MLP(8)
        fully_shard(model)
        with self.assertRaisesRegex(AssertionError, "FSDP does not support deepcopy"):
            copy.deepcopy(model)


if __name__ == "__main__":
    run_tests()



from typing import Union

import torch
import torch.utils._pytree as pytree
from torch.export.exported_program import ExportedProgram


__all__ = ["move_to_device_pass"]


def move_to_device_pass(
    ep: ExportedProgram, location: torch.device | str | dict[str, str]
) -> ExportedProgram:
    """
    Move the exported program to the given device.

    Args:
        ep (ExportedProgram): The exported program to move.
        location (Union[torch.device, str, Dict[str, str]]): The device to move the exported program to.
            If a string, it is interpreted as a device name.
            If a dict, it is interpreted as a mapping from
            the existing device to the intended one

    Returns:
        ExportedProgram: The moved exported program.
    """

    def _get_new_device(
        curr_device: torch.device,
        location: torch.device | str | dict[str, str],
    ) -> str:
        if isinstance(location, dict):
            if str(curr_device) in location:
                return location[str(curr_device)]
            else:
                return str(curr_device)
        else:
            return str(location)

    # move all the state_dict
    for k, v in ep.state_dict.items():
        if isinstance(v, torch.nn.Parameter):
            ep._state_dict[k] = torch.nn.Parameter(
                v.to(_get_new_device(v.device, location)),
                v.requires_grad,
            )
        else:
            ep._state_dict[k] = v.to(_get_new_device(v.device, location))

    # move all the constants
    for k, v in ep.constants.items():
        if isinstance(v, torch.Tensor):
            ep._constants[k] = v.to(_get_new_device(v.device, location))

    # move example_inputs if they exist
    if ep.example_inputs is not None:
        args, kwargs = ep.example_inputs
        moved_args = pytree.tree_map_only(
            torch.Tensor,
            lambda tensor: tensor.to(_get_new_device(tensor.device, location)),
            args,
        )
        moved_kwargs = pytree.tree_map_only(
            torch.Tensor,
            lambda tensor: tensor.to(_get_new_device(tensor.device, location)),
            kwargs,
        )
        ep._example_inputs = (moved_args, moved_kwargs)

    for m in ep.graph_module.modules():
        if isinstance(m, torch.fx.GraphModule):
            for node in m.graph.nodes:
                # move all the nodes kwargs with burnt-in device
                if "device" in node.kwargs:
                    kwargs = node.kwargs.copy()
                    kwargs["device"] = _get_new_device(kwargs["device"], location)
                    node.kwargs = kwargs

                if (
                    node.op == "call_function"
                    and node.target is torch.ops.aten.to.device
                ):
                    args = list(node.args)
                    # pyrefly: ignore [unsupported-operation]
                    args[1] = _get_new_device(args[1], location)
                    node.args = tuple(args)

                # move all the tensor metadata
                node.meta["val"] = pytree.tree_map(
                    lambda v: v.to(_get_new_device(v.device, location))
                    if isinstance(v, torch.Tensor)
                    else v,
                    node.meta.get("val"),
                )

    ep.validate()
    return ep



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# Keep a registry of all convolution operators.
import abc

import mslk.conv  # noqa: F401

import torch

from mslk.bench.common.utils import BenchOptions, do_bench
from mslk.quantize.triton.fp8_quantize import quantize_fp8_tensor


conv_op_registry = []


class ConvOpBase(metaclass=abc.ABCMeta):
    """Helper abstract class to define expected methods of conv ops."""

    @abc.abstractmethod
    def quantize(self, *args):
        """Function which quantizes inputs."""
        pass

    @abc.abstractmethod
    def compute(self, *args, **kwargs):
        """Function which performs main compute operation."""
        pass

    @abc.abstractmethod
    def quantize_and_compute(self, *args, **kwargs):
        """Function which quantizes inputs and performs main compute operation."""
        pass

    def preprocess(self, *args):
        """Preprocess inputs before benchmarking. These outputs will be passed to quantize."""
        return args

    def benchmark(
        self,
        *args,
        opts: BenchOptions,
        bench_quantize: bool,
    ) -> float:
        """Benchmark runtime of this operator."""
        return do_bench(
            lambda *a: self.quantize_and_compute(*a)
            if bench_quantize
            else self.compute(*a),
            args,
            opts,
        )

    @abc.abstractproperty
    def name(self) -> str:
        """Name of the operator."""
        pass

    @abc.abstractproperty
    def hip(self) -> bool:
        """Whether this operator supports AMD or not."""
        pass

    @abc.abstractproperty
    def cuda(self) -> bool:
        """Whether this operator supports Nvidia or not."""
        pass

    @property
    def supported(self) -> bool:
        """Whether this op will run on the current device."""
        if torch.version.hip is not None:
            return self.hip
        elif torch.version.cuda is not None:
            return self.cuda
        else:
            return False


def register_conv_op(op):
    """Decorator function for assembling all conv ops."""
    conv_op_registry.append(op())
    return op


def get_conv_ops() -> list[ConvOpBase]:
    """Get all registered conv ops."""
    return conv_op_registry


@register_conv_op
class TorchBaseline(ConvOpBase):
    """
    PyTorch baseline convolution.
    """

    def __init__(self):
        self.torch_compile = False

    def quantize(self, activation, filter, padding, stride, dilation):
        return (
            activation.to(torch.bfloat16),
            filter.to(torch.bfloat16),
            padding,
            stride,
            dilation,
        )

    def compute(self, activation, filter, padding, stride, dilation):
        if self.torch_compile:
            f = torch.compile(
                torch.nn.functional.conv3d,
                options={
                    "max_autotune": True,
                    "max_autotune_gemm_backends": "TRITON,CK,CUTLASS,ATEN",
                },
            )
        else:
            f = torch.nn.functional.conv3d

        return f(
            activation,
            filter,
            bias=None,
            stride=stride,
            padding=padding,
            dilation=dilation,
        )

    def quantize_and_compute(self, activation, filter, padding, stride, dilation):
        return self.compute(
            *self.quantize(activation, filter, padding, stride, dilation)
        )

    @property
    def name(self) -> str:
        return "torch_baseline"

    @property
    def hip(self) -> bool:
        return True

    @property
    def cuda(self) -> bool:
        return True


@register_conv_op
class F8F8BF16Conv(ConvOpBase):
    """
    FP8 convolution with rowwise scaling.
    """

    def preprocess(self, activation, filter, padding, stride, dilation):
        # Inputs and filters are provided in channels first layout.
        # Cutlass kernels support this but require the underlying memory
        # to be channels last. Torch enables this through the memory format
        # transformation which we assume has been applied ahead of time.
        activation = activation.to(memory_format=torch.channels_last_3d)
        filter = filter.to(memory_format=torch.channels_last_3d)
        return activation, filter, padding, stride, dilation

    def _quantize_tensor(self, x):
        """Quantize tensor to FP8 with rowwise scaling."""
        xq, x_scale = quantize_fp8_tensor(x)
        return xq, x_scale

    def quantize(self, activation, filter, padding, stride, dilation):
        # Quantize both input tensors
        activation_q, activation_scale = self._quantize_tensor(activation)
        filter_q, filter_scale = self._quantize_tensor(filter)

        # Compute combined scale for output
        # For conv, we need a single scale value
        scale = torch.tensor(
            [activation_scale * filter_scale],
            device=activation.device,
            dtype=torch.float32,
        )

        return activation_q, filter_q, scale, padding, stride, dilation

    def compute(self, activation_q, filter_q, scale, padding, stride, dilation):
        output = torch.ops.mslk.f8f8bf16_conv(
            activation_q,
            filter_q,
            scale,
            padding,
            stride,
            dilation,
        )
        return output

    def quantize_and_compute(self, activation, filter, padding, stride, dilation):
        activation_q, filter_q, scale, padding, stride, dilation = self.quantize(
            activation, filter, padding, stride, dilation
        )
        return self.compute(activation_q, filter_q, scale, padding, stride, dilation)

    @property
    def name(self) -> str:
        return "f8f8bf16_conv"

    @property
    def hip(self) -> bool:
        # Currently only supported on CUDA
        return False

    @property
    def cuda(self) -> bool:
        return True
